@@ -53,16 +53,17 @@
  * Now, how long shall be cached all handles? As long
  * as there are users. Each open connection is a user.
  * Additionally every "driver"-object is a user.
- * The drivier-object is allocated in each thread
+ * The driver-object is allocated in each thread
  * that tries to connect to DB and is released when
  * thread ends.
  }}} */
 #include "Oracle.h"
+
 DBISTATE_DECLARE;
 
 typedef struct llist_t llist_t;
 /* small implementation for doubly-linked list {{{ */
-struct llist_t{
+struct llist_t {
     llist_t * left;
     llist_t * right;
 };
@@ -74,6 +75,8 @@ struct llist_t{
     list->right = list->left = list;\
 }while(0)
 
+#define llist_is_initialized(list) ((list)->left != NULL)
+
 #define llist_add(aleft, aright) do{\
     llist_t * old;\
     llist_t * left = aleft;\
@@ -87,12 +90,23 @@ struct llist_t{
 
 #define llist_drop(ael) do{\
     llist_t * el = ael;\
-    if(llist_empty(el)) return;\
-    el->left->right = el->right;\
-    el->right->left = el->left;\
-    llist_init(el);\
+    if(!llist_empty(el)) {\
+        el->left->right = el->right;\
+        el->right->left = el->left;\
+        llist_init(el);\
+    } \
 }while(0)
 
+#define llist_size(list) ({\
+    llist_t * base = list;\
+    int size = 0;\
+    while(base->left != list) {\
+        size++;\
+        base = base->left;\
+    }\
+    size;\
+})
+
 // this is pointer to the left element in chain
 #define llist_left(list) (list)->left
 
@@ -117,6 +131,14 @@ dbd_dr_globals_init()
     dTHX;
     MUTEX_INIT(&mng_lock);
 #endif
+    // Ancient-Wizard:
+    // I believe this list, being static, was intented to be shared by all threads!
+    //  Additional comments in the code suggest this as well.
+    // However as it was written every thread was nuking the setup of the last
+    //  startng with a fresh empty list. Smells like a memory leak at best
+    //  and a crash at worst. We received both.
+    if ( llist_is_initialized(&mng_list) ) return;
+
     llist_init(&mng_list);
     dr_instances = 0;
     mng_env = NULL;
@@ -124,9 +146,9 @@ dbd_dr_globals_init()
 }
 /*}}}*/
 
-struct box_st{
+struct box_st {
     llist_t lock;
-    int refs; /* this shall be positiv for OCIEnv and negativ for Session Pool */
+    int refs; /* this shall be positive for OCIEnv and negativ for Session Pool */
 };
 
 struct env_box_st
@@ -311,7 +333,7 @@ simple_connect(pTHX_ dblogin_info_t * ctrl)
     sword status;
     ub4 ulen, plen, ctype;
 
-    tracer(imp_dbh, 3, 3, "using OCIEnv %p to connect\n", imp_dbh->envhp);
+    tracer(imp_dbh, 3, 3, "using OCIEnv envhp:%p to connect\n", imp_dbh->envhp);
 
     OCIHandleAlloc_ok(
             imp_dbh, imp_dbh->envhp, &imp_dbh->errhp,
@@ -895,7 +917,7 @@ int cnx_establish
 (pTHX_ dblogin_info_t * ctrl)
 {
     imp_dbh_t * imp_dbh = ctrl->imp_dbh;
-    env_box_t * env_box;
+    env_box_t * env_box = NULL;
 
     if(!figure_out_charsets(aTHX_ ctrl)) return FALSE;
     /* try to find existing OCIEnv */
@@ -956,8 +978,15 @@ dbd_dr_mng()
     }
     if(mng_env == NULL)
     {
+        // Now that this env handle is actually shared by threads the safest setup uses OCI_THREADED
+        ub4 oci_mode = OCI_DEFAULT;
+        // I haven't been able to prove that this is needed (TBD)
+        // I'm still chasing another SEGV in an unknown call as of yet!
+        // #if defined(USE_ITHREADS)
+        // oci_mode |= OCI_THREADED;
+        // #endif
         sword status = OCIEnvNlsCreate(
-                &mng_env, OCI_DEFAULT,
+                &mng_env, oci_mode,
                 0, NULL, NULL, NULL,
                 0, NULL, 0, 0
         );
@@ -1010,10 +1039,41 @@ cnx_detach(pTHX_ imp_dbh_t * imp_dbh)
     }
     else {
 #endif
-        (void)OCISessionEnd(
-                imp_dbh->svchp, imp_dbh->errhp, imp_dbh->seshp, OCI_DEFAULT
-        );
-        (void)OCIServerDetach(imp_dbh->srvhp, imp_dbh->errhp, OCI_DEFAULT);
+
+        sword status = OCI_SUCCESS;
+
+        // GLOBAL Perl END's may cause these to SEGV due to filehandles (just a guess).
+        //  Also forking can cause duplicate closings of a session. A preexisting handle in
+        //  the parent is duplicated in the child. The first process to call SessionEnd
+        //  wins, each successive call returns OCI_ERROR.
+        //
+        // The prior author included a comment about errors but was not specific and choose to ignore them.
+        //
+        // I thought I was being clever by adding the logging to these OCI calls;
+        //  to help debug the original SEGV issue. Best guess is the logging relies on
+        //  the filehandles being valid and they are not always valid, hence the SEGV
+        //  within Perl's IO printing.
+        //
+        // The OCI calls are successful but the logging is not. I don't like silent failures
+        //  so we'll use a more direct approach to handle this; keeping Perl and it's state
+        //  out of the picture.
+
+        // OCISessionEnd_log_stat( imp_dbh, imp_dbh->svchp, imp_dbh->errhp, imp_dbh->seshp, OCI_DEFAULT, status );
+        // OCIServerDetach_log_stat( imp_dbh, imp_dbh->srvhp, imp_dbh->errhp, OCI_DEFAULT, status );
+
+        if (( status = OCISessionEnd( imp_dbh->svchp, imp_dbh->errhp, imp_dbh->seshp, OCI_DEFAULT )) != OCI_SUCCESS
+           && status != OCI_ERROR )
+        {
+            // by printing the pointers I could see the children and parent were
+            //  using the same handles. The first is successful, the rest fail. (seems reasonable enough)
+            fprintf( stderr, "OCISessionEnd() failed: %d %p %p %p pid=%d\n", status, imp_dbh->svchp, imp_dbh->errhp, imp_dbh->seshp, getpid());
+        }
+
+        if (( status = OCIServerDetach( imp_dbh->srvhp, imp_dbh->errhp, OCI_DEFAULT )) != OCI_SUCCESS )
+        {
+            fprintf( stderr, "OCIServerDetach() failed: %d pid=%d\n", status, getpid());
+        }
+
 #ifdef ORA_OCI_112
     }
 #endif
